import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from llmevalkit.function_calling.pipeline.semantic_checker import \
    SemanticChecker
from llmevalkit.function_calling.pipeline.static_checker import evaluate_static
from llmevalkit.function_calling.pipeline.types import (FunctionCallInput,
                                                        FunctionCallMetric,
                                                        PipelineResult,
                                                        SemanticResult,
                                                        StaticMetricResult,
                                                        StaticResult, ToolCall,
                                                        ToolSpec)
from llmevalkit.llm import LLMClient

# Default metric JSONL locations
_METRICS_DIR = Path(__file__).parent.parent / "metrics"
_DEFAULT_GENERAL = _METRICS_DIR / "function_call" / "general_metrics.jsonl"
_DEFAULT_FUNCSEL = (
    _METRICS_DIR / "function_selection" / "function_selection_metrics.jsonl"
)
_DEFAULT_PARAM = _METRICS_DIR / "parameter" / "parameter_metrics.jsonl"


class ReflectionPipeline:
    """
    High-level orchestration for function-call reflection.

    Modes:
      • static_only: schema checks
      • semantic_only: LLM metrics + transforms
      • run: full static → semantic → assemble → PipelineResult

    Supports sync & async, custom JSONL overrides, and any registered LLM.
    """

    def __init__(
        self,
        metrics_client: LLMClient,
        codegen_client: Optional[LLMClient] = None,
        general_metrics: Optional[
            Union[Path, List[FunctionCallMetric]]
        ] = _DEFAULT_GENERAL,
        function_metrics: Optional[
            Union[Path, List[FunctionCallMetric]]
        ] = _DEFAULT_FUNCSEL,
        parameter_metrics: Optional[
            Union[Path, List[FunctionCallMetric]]
        ] = _DEFAULT_PARAM,
        transform_enabled: Optional[bool] = False,
    ):

        self.metrics_client = metrics_client
        self.codegen_client = codegen_client
        if codegen_client is None:
            codegen_client = metrics_client

        self.general_metrics = general_metrics
        self.function_metrics = function_metrics
        self.parameter_metrics = parameter_metrics

        metrics_definitions = []

        for metrics in (
            self.general_metrics,
            self.function_metrics,
            self.parameter_metrics,
        ):
            if not metrics:
                metrics_definitions.append(None)
                continue
            if not isinstance(metrics, (Path, list)):
                raise TypeError(
                    "metrics must be Path or List[FunctionCallMetric] or None"
                )
            if isinstance(metrics, list) and all(
                isinstance(x, FunctionCallMetric) for x in metrics
            ):
                metrics_definitions.append(
                    [metric.model_dump() for metric in metrics])
            else:
                if not metrics.is_file():
                    raise FileNotFoundError(
                        f"Metrics file not found: {metrics}")
                metrics_definitions.append(
                    [
                        json.loads(json_obj)
                        for json_obj in metrics.read_text(encoding="utf8").splitlines()
                        if json_obj.strip()
                    ]
                )

        gen_defs, fun_defs, par_defs = None, None, None

        if metrics_definitions:
            gen_defs = metrics_definitions[0]
            if len(metrics_definitions) >= 2:
                fun_defs = metrics_definitions[1]
                if len(metrics_definitions) >= 3:
                    par_defs = metrics_definitions[2]

        # 3) Initialize semantic checker
        self.semantic_checker = SemanticChecker(
            general_metrics=gen_defs,
            function_metrics=fun_defs,
            parameter_metrics=par_defs,
            metrics_client=self.metrics_client,
            codegen_client=self.codegen_client,
            transform_enabled=transform_enabled,
        )

    @staticmethod
    def static_only(
        inventory: List[ToolSpec],
        call: ToolCall,
    ) -> StaticResult:
        """
        Run schema-based static checks.

        Returns:
            StaticResult with per-check results and final_decision.
        """
        try:
            return evaluate_static(inventory, call)
        except Exception as e:
            return StaticResult(
                metrics={
                    "json_schema_validation": StaticMetricResult(
                        description="Invalid JSON schema",
                        valid=False,
                        explanation=f"error parsing JSON schema: {str(e)}",
                    )
                },
                final_decision=False,
            )

    def semantic_sync(
        self,
        conversation: Union[str, List[Dict[str, str]]],
        inventory: List[ToolSpec],
        call: ToolCall,
        retries: Optional[int] = 2,
        transform_enabled: Optional[bool] = None,
    ) -> SemanticResult:
        """
        Synchronous LLM-based semantic metrics (+ optional transforms).
        """
        # delegate to SemanticChecker
        return self.semantic_checker.run_sync(
            inventory,
            call,
            conversation,
            retries=retries,
            transform_enabled=transform_enabled,
        )

    async def semantic_async(
        self,
        conversation: Union[str, List[Dict[str, str]]],
        inventory: List[ToolSpec],
        call: ToolCall,
        retries: Optional[int] = 2,
        max_parallel: Optional[int] = 10,
        transform_enabled: Optional[bool] = None,
    ) -> SemanticResult:
        """
        Async LLM-based semantic metrics (+ optional transforms).
        """
        return await self.semantic_checker.run_async(
            inventory,
            call,
            conversation,
            retries=retries,
            max_parallel=max_parallel,
            transform_enabled=transform_enabled,
        )

    def run_sync(
        self,
        conversation: Union[str, List[Dict[str, str]]],
        inventory: List[ToolSpec],
        call: ToolCall,
        continue_on_static: Optional[bool] = False,
        retries: Optional[int] = 1,
        transform_enabled: Optional[bool] = None,
    ) -> PipelineResult:
        """
        Full sync pipeline: static -> semantic -> assemble PipelineResult.
        """
        static_res = self.static_only(inventory, call)

        if not static_res.final_decision and not continue_on_static:
            inputs = FunctionCallInput(
                conversation_context=conversation,
                tools_inventory=inventory,
                tool_call=call,
            )
            return PipelineResult(
                inputs=inputs,
                static=static_res,
                semantic=SemanticResult(
                    general=None,
                    function_selection=None,
                    parameter=None,
                    transform=None,
                ),
                overall_valid=False,
            )

        semantic_res = self.semantic_sync(
            conversation, inventory, call, retries, transform_enabled
        )
        return PipelineResult(
            inputs=FunctionCallInput(
                conversation_context=conversation,
                tools_inventory=inventory,
                tool_call=call,
            ),
            static=static_res,
            semantic=semantic_res,
            overall_valid=True,
        )

    async def run_async(
        self,
        conversation: Union[str, List[Dict[str, str]]],
        inventory: List[ToolSpec],
        call: ToolCall,
        continue_on_static: Optional[bool] = False,
        retries: Optional[int] = 1,
        max_parallel: Optional[int] = 10,
        transform_enabled: Optional[bool] = None,
    ) -> PipelineResult:
        """
        Full async pipeline: static -> semantic -> assemble PipelineResult.
        """
        static_res = self.static_only(inventory, call)

        if not static_res.final_decision and not continue_on_static:
            inputs = FunctionCallInput(
                conversation_context=conversation,
                tools_inventory=inventory,
                tool_call=call,
            )
            return PipelineResult(
                inputs=inputs,
                static=static_res,
                semantic=SemanticResult(
                    general=None,
                    function_selection=None,
                    parameter=None,
                    transform=None,
                ),
                overall_valid=False,
            )

        semantic_res = await self.semantic_async(
            conversation, inventory, call, retries, max_parallel, transform_enabled
        )
        return PipelineResult(
            inputs=FunctionCallInput(
                conversation_context=conversation,
                tools_inventory=inventory,
                tool_call=call,
            ),
            static=static_res,
            semantic=semantic_res,
            overall_valid=True,
        )
